#!/usr/bin/env python
"""
:Author Patrik Valkovic
:Created 11.12.2024 16:18
:Licence MIT
Part of grammpy

"""
from inspect import isclass
from itertools import chain
from typing import TYPE_CHECKING, Union, Any, List, Iterable, Type, Set, cast
from ... import Terminal, END_OF_INPUT, Rule, Nonterminal, EPSILON
from ...exceptions import ParsingTableDiscrepancyException, NonterminalIsMissingException, NoRuleForLookAheadException, \
    ParsingAmbiguityException, NotRuleException

if TYPE_CHECKING:  # pragma: no cover
    from .create_ll_parsing_table import LLTableType


class _StackTerminalWrapper:
    def __init__(self, symbol, parent_rule):
        self.symbol = symbol
        self.parent_rule = parent_rule


def ll(start_nonterminal, sequence, parsing_table, look_ahead, *, raise_on_ambiguity=True):
    # type: (Type[Nonterminal], Iterable[Union[Terminal, Any]], LLTableType, int, Any, bool) -> Nonterminal
    """
    Parse input sequence using look-ahead of LL(k) parsing.
    :param start_nonterminal: Starting nonterminal of grammar, also the type of root node of the parsed AST.
    :param sequence: Sequence to parse. Must be iterator and support next() method.
    :param parsing_table: Parsing table to use for parsing. Look ahead for the table must be the same as look_ahead parameter.
    :param look_ahead: Look-ahead to use for parsing.
    :param raise_on_ambiguity: Raise exception if there are ambiguous rules applicable for current nonterminal and look ahead.
    :return: Root node of the parsed AST.
    """
    stack = []  # type: List[_StackTerminalWrapper]
    ending_sequence = tuple([END_OF_INPUT] * look_ahead)
    start = start_nonterminal()
    stack.append(_StackTerminalWrapper(start, None))

    input_sequence = chain(sequence, ending_sequence)
    current_lookahead = tuple([next(input_sequence) for _ in range(look_ahead)])
    current_symbol_index = 0
    while True:
        if current_lookahead == ending_sequence and len(stack) == 0:
            return start
        if len(current_lookahead) < look_ahead:
            current_lookahead = tuple([*current_lookahead, next(input_sequence)])
            continue
        top_stack = stack.pop()
        if top_stack.symbol is EPSILON:
            eps_terminal = Terminal(EPSILON)
            top_stack.parent_rule._to_symbols.append(eps_terminal)
            eps_terminal._set_from_rule(top_stack.parent_rule)
            continue
        if not isinstance(top_stack.symbol, Nonterminal):
            if hash(top_stack.symbol) != hash(current_lookahead[0]):
                raise ParsingTableDiscrepancyException(current_symbol_index, f"Expected terminal {top_stack.symbol}, but found {current_lookahead[0]}")
            term_symbol = current_lookahead[0]
            if not isinstance(term_symbol, Terminal):
                term_symbol = Terminal(term_symbol)
            current_lookahead = tuple(current_lookahead[1:])
            current_symbol_index += 1
            top_stack.parent_rule._to_symbols.append(term_symbol)
            term_symbol._set_from_rule(top_stack.parent_rule)
            continue
        # Connect the nonterminal
        if top_stack.parent_rule is not None:
            top_stack.parent_rule._to_symbols.append(top_stack.symbol)
            top_stack.symbol._set_from_rule(top_stack.parent_rule)
        # Find rule to use
        nonterminal_instance = top_stack.symbol
        nonterminal = type(nonterminal_instance)
        if nonterminal not in parsing_table:
            raise NonterminalIsMissingException(current_symbol_index, nonterminal, f"There are no rules for nonterminal {nonterminal.__name__}")
        table_for_nonterminal = parsing_table[nonterminal]
        if current_lookahead not in table_for_nonterminal:
            raise NoRuleForLookAheadException(current_symbol_index, f"No rule for {nonterminal.__name__} with lookahead {current_lookahead} found")
        rule_to_use = table_for_nonterminal[current_lookahead]
        if isinstance(rule_to_use, set) and len(rule_to_use) == 0:
            raise NoRuleForLookAheadException(current_symbol_index, f"No rule for {nonterminal.__name__} with lookahead {current_lookahead} found")
        if isinstance(rule_to_use, set) and len(rule_to_use) > 1 and raise_on_ambiguity:
            raise ParsingAmbiguityException(current_symbol_index, [*rule_to_use], f"Ambiguity found for {nonterminal.__name__} and lookahead {current_lookahead}")
        if isinstance(rule_to_use, set):
            rule_to_use = next(iter(rule_to_use))
        if not isclass(rule_to_use) or not issubclass(rule_to_use, Rule):
            raise NotRuleException(rule_to_use)
        # Apply the rule
        rule_instance = rule_to_use()
        nonterminal_instance._set_to_rule(rule_instance)
        rule_instance._from_symbols.append(nonterminal_instance)
        for symbol in reversed(rule_instance.right):
            if symbol == EPSILON:
                stack.append(_StackTerminalWrapper(EPSILON, rule_instance))
            elif isclass(symbol) and issubclass(symbol, Nonterminal):
                stack.append(_StackTerminalWrapper(symbol(), rule_instance))
            else:
                stack.append(_StackTerminalWrapper(symbol, rule_instance))



#!/usr/bin/env python
"""
:Author Patrik Valkovic
:Created 01.08.2017 07:52
:Licence MIT
Part of grammpy

"""

from typing import TYPE_CHECKING
if TYPE_CHECKING:  # pragma no cover
    from typing import TypeAlias

class _Epsilon:
    def __repr__(self):
        return 'EPSILON'
    def __str__(self):
        return 'EPSILON'
    def __reduce__(self):
        return _get_epsilon_instance, ()

EPSILON = _Epsilon()
EPS = EPSILON

EPSILON_TYPE = type(EPSILON)  # type: TypeAlias

def _get_epsilon_instance():
    return EPSILON


class _EndOfInput:
    def __repr__(self):
        return 'END_OF_INPUT'
    def __str__(self):
        return 'END_OF_INPUT'
    def __reduce__(self):
        return _get_end_of_input_instance, ()

END_OF_INPUT = _EndOfInput()
EOI = END_OF_INPUT

END_OF_INPUT_TYPE = type(END_OF_INPUT)  # type: TypeAlias

def _get_end_of_input_instance():
    return END_OF_INPUT



#!/usr/bin/env python
"""
:Author Patrik Valkovic
:Created 04.12.2024 11:53
:Licence MIT
Part of grammpy

"""
from inspect import isclass
from typing import TYPE_CHECKING, Dict, Set, Tuple, Union, Type, Any
from grammpy import END_OF_INPUT, Nonterminal, EPSILON

if TYPE_CHECKING:   # pragma: no cover
    from grammpy import Grammar, Terminal, Rule, END_OF_INPUT_TYPE
    from ..FirstTable.create_first_table import FirstTableType
    FollowTableTypeValue = Set[
        Union[
            Tuple[
                Union[
                    Type[Terminal],
                    Any,
                    END_OF_INPUT_TYPE
                ],
                ...
            ]
        ]
    ]
    FollowTableType = Dict[
        Union[Type[Nonterminal]],
        FollowTableTypeValue
    ]

def create_follow_table(grammar, first_table, look_ahead):
    # type: (Grammar, FirstTableType, int) -> FollowTableType
    """
    Given LL(k) grammar and its corresponding first table, create follow table.
    :param grammar: Grammar for which follow table is created.
    :param first_table: First table for the grammar.
    :param look_ahead: Look ahead of the parser, must be same or lower as the first table.
    :return: Follow table for the grammar
    """
    table = {nt: set() for nt in grammar.nonterminals}  # type: FollowTableType
    eoi_tuple = tuple([END_OF_INPUT]*look_ahead)
    table[grammar.start].add(eoi_tuple)
    updated = True

    def add_from_set(from_nonterminal, to_symbols_set):
        # type: (Type[Nonterminal], FollowTableTypeValue) -> None
        nonlocal updated
        original_size = len(table[from_nonterminal])
        for symbols in to_symbols_set:
            if len(symbols) == look_ahead:
                table[from_nonterminal].add(symbols)
        if len(table[from_nonterminal]) > original_size:
            updated = True

    while updated:
        updated = False
        for rule in grammar.rules:  # type: Type[Rule]
            left = rule.fromSymbol
            right = rule.right
            right_side_generates_epsilon = True
            right_side_first = table[left].copy()
            for i in range(len(right)-1, -1, -1):
                symbol = right[i]
                # check if it is terminal
                if symbol == EPSILON:
                    add_from_set(left, right_side_first)
                    continue
                if not isclass(symbol) or not issubclass(symbol, Nonterminal):
                    right_side_first = {
                        tuple([symbol, *k][:look_ahead]) for k in right_side_first
                    }
                    if right_side_generates_epsilon:
                        right_side_first.add(tuple([symbol]))
                    right_side_generates_epsilon = False
                    continue
                # it is nonterminal
                if right_side_generates_epsilon:
                    add_from_set(symbol, table[left])
                add_from_set(symbol, right_side_first)
                right_side_first = {
                    tuple([*c ,*e][:look_ahead]) for c in first_table[symbol] if c != EPSILON for e in right_side_first
                }
                right_side_generates_epsilon = right_side_generates_epsilon and EPSILON in first_table[symbol]

    return table



#!/usr/bin/env python
"""
:Author Patrik Valkovic
:Created 20.12.2024 17:36
:Licence MIT
Part of grammpy

"""
from .LLParsingException import LLParsingException

class ParsingTableDiscrepancyException(LLParsingException):
    """
    Exception raised when the nonterminal on the stack doesn't match the nonterminal in the input sequence.
    This should never happen and in general points to discrepancy between the parsing table and the rule.
    The rule probably doesn't generate the terminal that is in the look-ahead sequence.
    """
    pass



#!/usr/bin/env python
"""
:Author Patrik Valkovic
:Created 20.12.2024 17:43
:Licence MIT
Part of grammpy

"""
from typing import Any, TYPE_CHECKING, Type
from .LLParsingException import LLParsingException

if TYPE_CHECKING:  # pragma: no cover
    from .. import Nonterminal

class NonterminalIsMissingException(LLParsingException):
    """
    Exception that occurs when there is nonterminal on top of the stack that has no entry in the parsing table.
    This should not happen in general and suggest there are some problems with the grammar.
    """

    def __init__(self, position, nonterminal, *args):
        # type: (int, Type[Nonterminal], Any) -> None
        self.nonterminal = nonterminal
        super().__init__(position, *args)



#!/usr/bin/env python
"""
:Author Patrik Valkovic
:Created 04.12.2024 11:51
:Licence MIT
Part of grammpy

"""

from typing import TYPE_CHECKING
from .create_follow_table import create_follow_table

if TYPE_CHECKING:  # pragma: no cover
    from .create_follow_table import FollowTableType, FollowTableTypeValue


#!/usr/bin/env python
"""
:Author Patrik Valkovic
:Created 01.12.2024 17:56
:Licence MIT
Part of grammpy

"""
from inspect import isclass
from typing import TYPE_CHECKING, Dict, List, Set, Union, Type, Tuple, Any
from grammpy import Nonterminal, Terminal, EPSILON
if TYPE_CHECKING:  # pragma: no cover
    from grammpy import Grammar, EPSILON_TYPE
    FirstTableTypeValue = Set[
        Union[
            EPSILON_TYPE,
            Tuple[Union[Type[Terminal], Any], ...]
        ]
    ]
    FirstTableType = Dict[
        Type[Nonterminal],
        FirstTableTypeValue
    ]

def create_first_table(grammar, look_head):
    # type: (Grammar, int) -> FirstTableType
    """
    Given LL(n) grammar creates first table
    :param grammar: Grammar to create first table for
    :param look_head: Number of symbols to look ahead
    :return: First table
    """
    if type(look_head) != int:
        raise TypeError('Look ahead must be integer')
    if look_head < 1:
        raise ValueError('Look ahead must be at least 1')

    table = {nt: set() for nt in grammar.nonterminals}  # type: FirstTableType

    updated = True
    while updated:
        updated = False
        for rule in grammar.rules:
            from_symbol = rule.fromSymbol  # type: Type[Nonterminal]
            to_symbols = rule.right  # type: List[Union[Type[Terminal], Terminal, EPSILON_TYPE, Type[Nonterminal]]]

            def add_to_table(current_firsts):
                nonlocal updated
                # Add generated sequences to the FIRST set of `from_symbol`
                original_size = len(table[from_symbol])
                if current_firsts == EPSILON:
                    table[from_symbol].add(current_firsts)
                else:
                    table[from_symbol].add(tuple(current_firsts[:look_head]))
                if len(table[from_symbol]) > original_size:
                    updated = True

            def add_to_table_set(current_firsts):
                for current_first in current_firsts:
                    add_to_table(current_first)

            def iterate_rule(current_firsts, to_look, contains_epsilon_prefix):
                # type: (Set[Tuple], List, bool) -> None

                # If some prefix is long enough, add it to the table and ignore it for further processing
                long_enough_current_firsts = [c for c in current_firsts if len(c) >= look_head]
                for current in long_enough_current_firsts:
                    add_to_table(current)
                # Focus only on short prefixes
                current_firsts = set([c for c in current_firsts if len(c) < look_head])
                # If there is nothing to look at, add the current FIRST set to the table
                if len(to_look) == 0:
                    if contains_epsilon_prefix:
                        add_to_table(EPSILON)
                    add_to_table_set(current_firsts)
                    return
                symbol = to_look[0]

                # Right side is epsilon, skip it and look at next symbol
                if symbol == EPSILON:
                    return iterate_rule(current_firsts, to_look[1:], contains_epsilon_prefix)
                # Right side has terminal, add it to the FIRST set
                elif not isclass(symbol) or not issubclass(symbol, Nonterminal):
                    new_first = set(
                        [tuple([*c, symbol]) for c in current_firsts] +
                        ([tuple([symbol])] if contains_epsilon_prefix else []) # If there is epsilon prefix, FIRST may start with the symbol
                    )
                    return iterate_rule(new_first, to_look[1:], False)
                # Right side has nonterminal, add its FIRST table to the current symbol
                elif issubclass(symbol, Nonterminal):
                    nonterminal_entries = table[symbol].copy()
                    for entry in nonterminal_entries:  # type: Tuple | EPSILON_TYPE
                        is_eps = entry == EPSILON
                        if is_eps:
                            iterate_rule(current_firsts, to_look[1:], contains_epsilon_prefix)
                        else:
                            new_first = set(
                                [tuple([*c, *entry]) for c in current_firsts] +
                                ([entry] if contains_epsilon_prefix else [])
                            )
                            iterate_rule(new_first, to_look[1:], False)
                else:
                    raise ValueError('Unknown symbol type, should never happen', symbol)  # pragma no cover

            iterate_rule(set(), to_symbols, True)

    return table




#!/usr/bin/env python
"""
:Author Patrik Valkovic
:Created 03.08.2017 10:04
:Licence MIT
Part of grammpy

"""
from typing import Any
from .GrammpyException import GrammpyException

class NotRuleException(GrammpyException, TypeError):
    """
    Passed something else than Rule class
    """
    def __init__(self, rule, *args):
        # type: (Any, Any) -> None
        super().__init__(*args)
        self.object = rule



#!/usr/bin/env python
"""
:Author Patrik Valkovic
:Created 05.12.2024 21:01
:Licence MIT
Part of grammpy

"""
from inspect import isclass
from typing import TYPE_CHECKING, Dict, Union, Type, Tuple, Set, Any
from grammpy import Nonterminal, EPSILON

if TYPE_CHECKING:   # pragma: no cover
    from grammpy import Grammar, Terminal, Rule, END_OF_INPUT_TYPE, EPSILON_TYPE
    from ...transforms.FirstTable.create_first_table import FirstTableType
    from ...transforms.FollowTable.create_follow_table import FollowTableType
    LLTableType = Dict[
        Type[Nonterminal],
        Dict[
            Tuple[
                Union[
                    END_OF_INPUT_TYPE,
                    Type[Terminal],
                    Any
                ],
                ...
            ],
            Union[
                Set[Type[Rule]],
                Type[Rule]
            ]
        ]
    ]

def create_ll_parsing_table(g, first, follow, look_ahead):
    # type: (Grammar, FirstTableType, FollowTableType, int) -> LLTableType
    """
    Create LL parser for given grammar.
    You must provide FIRST and FOLLOW table for the grammar with the proper look ahead.
    :param g: Grammar to create LL parser for.
    :param first: FIRST table for the grammar.
    :param follow: FOLLOW table for the grammar.
    :param look_ahead: Number of symbols that the parser will look.
    :return: LL(k) parsing table.
    """
    parsing_table = {N: dict() for N in g.nonterminals}

    for rule in g.rules:
        left = rule.fromSymbol
        right = rule.right
        # generate FIRST of the right side
        can_generate_epsilon = True
        first_of_right = set()
        for symbol in right:  # type: Union[Type[Terminal], Terminal, EPSILON_TYPE]
            if symbol == EPSILON:
                continue
            if not isclass(symbol) or not issubclass(symbol, Nonterminal):
                first_of_right = {
                    tuple([*e, symbol]) for e in first_of_right
                }
                if can_generate_epsilon:
                    first_of_right.add(tuple([symbol]))
                can_generate_epsilon = False
            else:
                first_without_epsilon = {f for f in first[symbol] if f != EPSILON}
                first_of_right = {
                    tuple([*e, *([*f] if f is not EPSILON else [])]) for f in first[symbol] for e in first_of_right
                }
                if can_generate_epsilon:
                    first_of_right.update(f for f in first_without_epsilon)
                can_generate_epsilon = can_generate_epsilon and EPSILON in first[symbol]
        # add follow
        first_of_right = {
            tuple([*e, *f]) for f in follow[left] for e in first_of_right
        }
        if can_generate_epsilon:
            first_of_right.update(follow[left])
        # shorten sequences
        first_of_right = {tuple(e[:look_ahead]) for e in first_of_right}
        # add to the parsing table
        for sequence in first_of_right:
            if sequence not in parsing_table[left]:
                parsing_table[left][sequence] = set()
            parsing_table[left][sequence].add(rule)

    return parsing_table



#!/usr/bin/env python
"""
:Author Patrik Valkovic
:Created 20.12.2024 17:15
:Licence MIT
Part of grammpy

"""
from typing import Any

from .ParsingException import ParsingException


class LLParsingException(ParsingException):
    """
    Exception raised when the LL parsing fails.
    In addition has `position` attribute specifying index of the symbol from input sequence that caused the error.
    """""

    def __init__(self, position, *args):
        # type: (int, Any) -> None
        self.position = position
        super().__init__(*args)

    def __str__(self):
        return f"LL parsing error on symbol at {self.position}: {super().__str__()}"



#!/usr/bin/env python
"""
:Author Patrik Valkovic
:Created 01.12.2024 17:55
:Licence MIT
Part of grammpy

"""

from typing import TYPE_CHECKING
from .create_first_table import create_first_table

if TYPE_CHECKING:  # pragma: no cover
    from .create_first_table import FirstTableType, FirstTableTypeValue



#!/usr/bin/env python
"""
:Author Patrik Valkovic
:Created 20.12.2024 17:56
:Licence MIT
Part of grammpy

"""
from .LLParsingException import LLParsingException

class NoRuleForLookAheadException(LLParsingException):
    """
    There is no rule to apply for the current nonterminal and look ahead.
    """
    pass


#!/usr/bin/env python
"""
:Author Patrik Valkovic
:Created 20.12.2024 18:01
:Licence MIT
Part of grammpy

"""
from typing import Any, Type, List, TYPE_CHECKING
from .LLParsingException import LLParsingException

if TYPE_CHECKING:  # pragma: no cover
    from .. import Rule

class ParsingAmbiguityException(LLParsingException):
    """
    There are multiple rules to apply for the current nonterminal and look ahead.
    """

    def __init__(self, position, rules, *args):
        # type: (int, List[Type[Rule]], Any) -> None
        self.rules = rules
        super().__init__(position, *args)

